body { 
     font-family: Verdana,Geneva,Arial,Sans-serif; 
	font-size: 12px; 
	background-color : #FFFFFF 
}

a         { color : #2222FF; text-decoration : underline; }
a:visited { color : #000080; }
a:hover   { color : #2222FF; text-decoration : none; }

a.missing_page:visited { color : Red; }
a.missing_page:link    { color : Red; }
a.missing_page:hover   { color : Red; text-decoration : none; }

.nwikilogo {
  color:#89d;
  font-weight:bold;
  font-size:2em;
  line-height:2em;
}

.error { color : Red; font-weight:bold; }

#akmenu a
{
        display: block;
        text-decoration: none;
}

/* Accesskey'd links -- first letter underlined */
a.ak {	text-decoration: none; }
a.ak:hover { color:#2222FF; text-decoration: underline;}
a.ak:focus { color:#2222FF; text-decoration: underline;}
a.ak:first-letter { color:#2222FF; text-decoration: underline; font-weight:bold; }

code { color : #000070 }
pre { color :  #000070 }

h4.entry { text-decoration: underline }
h1 {
  text-align: left;
  color:#AF4040;
}

h1.left_aligned {
  text-align: left; 
}

h4 {
  margin-bottom:0px;
  margin-top:0px;
}

h2 {color:#BF3030; }

html, body {
	margin: 0;
	padding: 0;
	}

#navbar {
	width: 120px;
	float: left;
	margin-left: -1px;
	margin-top: 10px;
	padding: 12px;
}
	
	
#content {
	padding: 10px;
	margin-left: 280px;
}
