(* Ocsigen
 * http://www.ocsigen.org
 * Copyright (C) 2011 Benedikt Becker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, with linking exception;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

val get_default_hostname : unit -> string
val get_default_port : unit -> int
val get_default_sslport : unit -> int
val default_protocol_is_https : unit -> bool
val get_default_links_xhr : unit -> bool

val debug_timings : bool ref

(** Not tracing by default. Can be dynamically set by adding ["#__trace"] to
    the URL. *)
val set_tracing : bool -> unit
val get_tracing : unit -> bool

(** Same as [Ocsigen_config.get_debugmode].
    On client side, returns [false] for now. *)
val get_debugmode : unit -> bool
