body {
  background-color: #fefefe;
  color: black;
  padding : 0px;
  margin-top: 0px;
  margin-bottom: 0px;
  margin-left: 30px;
  margin-right: 30px;
  text-align: justify;
  font-family: Verdana, Arial,sans-serif;
  font-size:0.9em;
}

div.principal {
  padding : 0px;
  margin: 10px;
}

h1 {
  padding-top : 1.5em;
  padding-bottom : 1.5em;
  margin: 0px;
  text-align: center;
}

h2 {
  margin: 0em 0px 0.5em 0px;
  padding: 1em 0.5em 0.2em 0.5em;
  font-size: 200%;
  font-weight: 700;
  clear: both;
  text-align: center;
}

h3 {
  margin: 0em 0px 0.5em 0px;
  padding: 0.2em 0.5em 0.2em 0.5em;
  font-size: 100%;
  font-weight: 700;
  background-color: #3860aa;
  background-image: url("bulles-bleues.png");
  background-position: bottom right;
  background-repeat: no-repeat;
  color: white;
  clear: both;
  text-align: center;
}

h4 {
  margin: 0em 0em 0.5em 0em;
  padding: 0.2em 0.5em 0.2em 0.5em;
  font-size: 100%;
  font-weight: 700;
  background-color: #3860aa;
  background-image: url("bulles-bleues.png");
  background-position: bottom right;
  background-repeat: no-repeat;
  color: white;
}

h5 {
  margin: 0em 0em 0.5em 0em;
  padding: 0em;
  font-size: 100%;
  font-weight: 700;
  color: #3860aa;
}

a {
  color: #ff5555;
  font-weight: bold;
}

a:hover {
  color: #ff8888;
}

a:visited {
  color: #ff5555;
}

p {
  margin: 0em 0em 0.5em 0em;
  padding: 0px;
}

ul {
  list-style-type: circle;
}

em {
  color: #882222;
}

.emph {
  font-style: italic;
  font-weight: 700;
  font-size: 100%;
}

.clear {
  clear: both;
  visibility: hidden;
}

.importantwarning {
  font-weight: 700;
  background-image: url("bulles-rouges-fond.png");
  background-position: bottom right;
  background-color: #9d1010;
  background-repeat: no-repeat;
  color: white;
  padding: 5px;
  margin: 10px 0px 10px 0px;
}

.importantwarning a {
  color: #ffff00;
}

.importantwarning a:visited {
  color: #ffff88;
}

.importantwarning h4 {
  color: #ffff00;
}

.importantwarning em {
  color: #ffcc88;
}

p.warning {
  font-weight: 700;
  color: #882222;
}

p.footer {
  clear: both;
  font-size: 70%;
  border-top: 1px solid #000000;
  margin: 0px 10px 0px 10px;
  padding: 5px 0px 0px 0px;
}
p.footer img {
  vertical-align: middle;
}

div.col2tiers {
  float:left;
  width: 65%;
  margin: 0px 0px 0px 0px;
  padding: 0px;
}

div.threecol1 {
  float:left;
  width: 48%;
  margin: 0px;
  padding: 0px;
}
div.threecol2 {
  float:right;
  width: 48%;
  margin: 0px;
  padding: 0px;
}
div.threecol3 {
  float:right;
  width: 32%;
  margin: 0px;
  padding: 0px;
}
div.onecol {
  padding: 1em 5px 1em 5px;
  margin-bottom: 1em;
}
div.twocol1 {
  float:left;
  width: 49%;
  margin: 0px 5px 15px 0px;
}
div.twocol2 {
  float:right;
  width: 49%;
  margin: 0px 0px 15px 5px;
}

div.menu {
  margin: 0px 0px 1em 0px;
}

ul.menu {
  color: #3860aa;
  text-align: right;
  padding: 0px 0px 0.2em 0px;
  margin: 0px;
}

ul.menuprincipal {
  border-bottom: 1px solid #000000;
  background-color: white;
}

.menu li {
  display: inline;
  border-right: solid 1px black;
  font-weight: bold;
  padding: 0px 5px;
  margin: 0px;
}
.menu li.last {
  border: none;
}
.sousmenu {
  font-size: 80%;
}
ul.sousmenu {
  margin: 2px 0px 0px 0px;
}

pre{
  background-color: #d0d9eb;
  margin: 0.5em 0em 1em 0em;
  padding: 5px;
  overflow:auto;
}

Div.camel {
  background-image: url("caml-bleu140.jpg");
  background-position: bottom right;
  background-repeat: no-repeat;
  background-color: #3860aa;
  margin-top: 2em;
  padding: 1em;
  padding-bottom: 100px;
}

.camel p {
  color: white;
  margin: 0em 0em 0.5em 0em;
}

.camel h4 {
  margin-top: 0em;
  padding: 0.5em;
  text-align: center;
  background-image: none;
}

.camel em {
  color: #ff5555;
  font-weight: bold;
}

div.news {
  background-position: bottom right;
  background-repeat: no-repeat;
  background-color: #d0d9eb;
  padding: 5px;
  margin: 0px 0px 5px 0px;
}

div#mainfeatures {
  background-position: bottom right;
  background-repeat: no-repeat;
  background-color: #d0d9eb;
  padding: 5px;
  margin: 0px 0px 5px 0px;
}

#mainfeatures ul {
  text-align: left;
}


.news h3 {
  margin-top: 0em;
}

#mainfeatures h3 {
  margin-top: 0em;
}

dl {
  margin: 0 0 0em 0em;
}

dd {
  margin: 0 0 0.5em 0.5em;
}

dt {
  margin: 0 0 0em 0em;
  font-weight: bold;
}

div.encadre {
  background-color: #3860aa;
  background-image: url("bulles-bleues-fond.png");
  background-position: bottom right;
  background-repeat: no-repeat;
  color: white;
  padding: 0px 10px 1em 10px;
  margin-bottom: 1em;
}

div.sanstitre {
  padding-top: 1em;
}

.encadre h4 {
  margin: 1.5em 0em 0.5em 0em;
  padding: 0.2em 0em 0.2em 0em;
  background-color: transparent;
  background-image: none;
}

.encadre pre {
  margin: 0.5px 5px 1em 5px;
  color: black;
}

.encadre em {
  color: #ff9999;
}

.encadre pre em {
  color: #882222;
}

.encadre p {
  margin: 0px 0px 0.5em 0px;
}

.encadre th {
  color: #3860aa;
  background-color: #b8d0ff;
  padding: 0.5em;
  text-align: center;
  border: 2px solid #3860aa;
}
.encadre th a {
  color: #ff5555;
}
th.row {
  text-align: right;
}
th.col {
  width: 33%;
}
.encadre td {
  background-color: #6890da;
  border: 2px solid #3860aa;
  padding: 0.5em;
}
.encadre td a {
  color: #ff8888;
}
.encadre td.empty {
  background-color: #3860aa;
}
dl.blue {
  background-color: #6890da;
  border: 2px solid #3860aa;
  padding: 0.5em;
}

.hidden_widget {
  border: 1px solid #99ffaa;
  margin: 1em;
}

.hidden_widget .overlay {
  background-color: #aaa;
}

.test_title {
}
.test_description {
  font-size: small;
}
.monospace {
  font-family: monospace;
}
.test_logger {
  font-family: monospace;
  font-size: x-small;
}
.test_logger .logging_line {
  display: block;
}
.thebutton {
  background: gray;
  border: 1px solid black;
  padding: 1em;
}
